"""
Concrete :class:`~.TrackerBase` subclass for FLD
"""

from ... import errors, utils
from ..base import TrackerBase
from . import config, rules
from .jobs import FldTrackerJobs

import logging  # isort:skip
_log = logging.getLogger(__name__)


class FldTracker(TrackerBase):
    name = 'fld'
    label = 'FLD'
    torrent_source_field = 'FLD'

    setup_howto_template = (
        '{howto.introduction}\n'
        '\n'
        '{howto.next_section}. Announce Passkey\n'
        '\n'
        '   {howto.current_section}.1 On the website, go to Settings -> Security\n'
        '       and copy your personal ANNOUNCE_KEY.\n'
        '   {howto.current_section}.2 $ upsies set trackers.{tracker.name}.announce_key ANNOUNCE_KEY\n'
        '\n'
        '{howto.next_section}. API Key\n'
        '\n'
        '   {howto.current_section}.1 On the website, go to Settings -> Security\n'
        '       and copy your personal API_KEY.\n'
        '   {howto.current_section}.2 $ upsies set trackers.{tracker.name}.apikey API_KEY\n'
        '\n'
        '{howto.screenshots}\n'
        '\n'
        '{howto.autoseed}\n'
        '\n'
        '{howto.reuse_torrents}\n'
        '\n'
        '{howto.upload}\n'
    )

    TrackerJobs = FldTrackerJobs
    TrackerConfig = config.FldTrackerConfig
    cli_arguments = config.cli_arguments
    rules = rules

    async def _login(self, *, tfa_otp=None):
        pass

    async def confirm_logged_in(self):
        pass

    async def _logout(self):
        pass

    async def get_announce_url(self):
        return '/'.join((
            self.options['announce_url'].rstrip('/'),
            self.options['announce_key'].get_secret_value(),
        ))

    async def upload(self, tracker_jobs):
        _log.debug('Uploading to %r', self.options['upload_url'])
        _log.debug('POSTing data:')
        post_data = tracker_jobs.post_data
        for k, v in post_data.items():
            _log.debug(' * %s = %s', k, v)

        files = {
            'meta_info': {
                'file': tracker_jobs.torrent_filepath,
                'mimetype': 'application/octet-stream',
            },
        }
        _log.debug('Files: %r', files)

        response = await utils.http.post(
            url=self.options['upload_url'],
            cache=False,
            user_agent=True,
            headers={'Authorization': f'Bearer {self.options["apikey"].get_secret_value()}'},
            files=files,
            data=post_data,
        )
        _log.debug('Upload response: %r', response)
        json = response.json()
        _log.debug('Upload response: %r', json)
        try:
            if json['success']:
                # Upload response: {
                #     'success': True,
                #     'message': "Success",
                #     'torrent_id': "<torrent id>",
                #     'torrent_url': "<torrent detail url>",
                #     'torrent_download': "<torrent download url>",
                # }
                return json['torrent_url']
            else:
                # Upload response: {
                #     'success': False,
                #     'message': "<error message>",
                # }
                raise errors.RequestError(f'Upload failed: {json["message"]}')
        except KeyError as e:
            raise RuntimeError(f'Unexpected response: {response}') from e



"""
Dummy image uploader for testing and debugging
"""

import asyncio
import os

from .. import errors, utils
from . import base

if utils.is_running_in_development_environment():

    class DummyImagehostConfig(base.ImagehostConfigBase):
        hostname: utils.config.fields.string(
            default='localhost',
            description='Host name in dummy image URLs.',
        )

    class DummyImagehost(base.ImagehostBase):
        """Dummy service for testing and debugging"""

        name = 'dummy'

        Config = DummyImagehostConfig

        description = (
            'This is a fake image hosting service '
            'that is used for testing and debugging.'
        )

        cli_arguments = {
            ('--hostname', '-n'): {
                'help': 'Host name in the dummy URL',
            },
        }

        async def _upload_image(self, image_path):
            try:
                utils.fs.assert_file_readable(image_path)
            except errors.ContentError as e:
                raise errors.RequestError(e) from e
            else:
                await asyncio.sleep(0.6)
                url = f'http://{self.config["hostname"]}/{os.path.basename(image_path)}'
                return url



"""
Create torrent file
"""

import functools
import os

from .. import btclients, errors, utils
from . import base

import logging  # isort:skip
_log = logging.getLogger(__name__)


class CreateTorrentJob(base.JobBase):
    """
    Create torrent file

    This job adds the following signals to :attr:`~.JobBase.signal`:

        ``announce_url``
            Emitted before and after successful announce URL retrieval.
            Registered callbacks get an :class:`Ellipsis` to indicate the
            retrieval attempt and the announce URL if the attempt was
            successful.

        ``file_tree``
            Emitted when torrent creation begins. Registered callbacks get
            nested `(file_name, file_size)` tuples (see
            :func:`~.utils.torrent.create`) as a positional argument. (See also
            :func:`.fs.format_file_tree`.)

        ``file_list``
            Emitted when torrent creation begins. Registered callbacks get a
            :class:`list` of paths to existing files that are in the torrent as
            a positional argument.

        ``progress_update``
            Emitted at roughly equal intervals to provide information about the
            torrent creation progress. Registered callbacks get a
            :class:`~.torrent.CreateTorrentProgress` or
            :class:`~.torrent.FindTorrentProgress` instance as a positional
            argument.
    """

    name = 'torrent'
    label = 'Torrent'
    cache_id = None

    def initialize(self, *, tracker, content_path, exclude_files=(), reuse_torrent_path=None):
        """
        Set internal state

        :param TrackerBase tracker: Return value of :func:`.trackers.tracker`
        :param content_path: Path to file or directory
        :param reuse_torrent_path: Path to existing torrent file to reuse piece
            hashes from (see :func:`~.utils.torrent.create`)
        :param exclude_files: Sequence of glob patterns (:class:`str`) and
            :class:`re.Pattern` objects (return value from :func:`re.compile`)
            that are matched against the relative path within the generated
            torrent

            Glob patterns are matched case-insensitively.

            .. note:: This sequence is combined with
                      :attr:`.TrackerBase.options`\\ ``["exclude"]``.
        """
        self._tracker = tracker
        self._content_path = content_path
        self._reuse_torrent_path = reuse_torrent_path
        self._torrent_path = os.path.join(
            self.home_directory,
            f'{utils.fs.basename(content_path)}.{tracker.name.lower()}.torrent',
        )

        self._exclude_files = list(self._tracker.options['exclude'])
        self._exclude_files.extend(exclude_files)

        self._torrent_process = None
        self._activity = ''

        self.signal.add('announce_url')
        self.signal.add('file_tree')
        self.signal.add('file_list')
        self.signal.add('progress_update')

    @property
    def activity(self):
        """
        What is currently being done

        ``"announce_url"``
            If we are getting the announce URL.

        ``"hashing"``
            If we are hashing files.

        ``"searching"``
            If we are searching for a torrent we can copy hashes from.

        ``"verifying"``
            If we are hash-checking a potentially matching torrent.

        ``""``
            None of the above is happening (e.g. we are getting the file tree).
        """
        return self._activity

    async def run(self):
        """Get announce URL from `tracker`, then execute torrent creation subprocess"""
        announce_url = await self._get_announce_url()
        if announce_url:
            self._start_torrent_process(announce_url)
            await self._torrent_process.join()

    async def _get_announce_url(self):
        self._activity = 'announce_url'
        self.signal.emit('announce_url', Ellipsis)
        try:
            announce_url = await self._tracker.get_announce_url()
        except errors.RequestError as e:
            self.error(e)
        else:
            self.signal.emit('announce_url', announce_url)
            return announce_url
        finally:
            self._activity = ''

    def _start_torrent_process(self, announce_url):
        self._torrent_process = utils.daemon.DaemonProcess(
            name=self.name,
            target=_torrent_process,
            kwargs={
                'content_path': self._content_path,
                'announce': announce_url,
                'source': self._tracker.torrent_source_field,
                'torrent_path': self._torrent_path,
                'exclude': self._exclude_files,
                'use_cache': not self.ignore_cache,
                'reuse_torrent_path': self._reuse_torrent_path,
                'piece_size_calculator': self._tracker.calculate_piece_size,
                'piece_size_min_max_calculator': self._tracker.calculate_piece_size_min_max,
            },
            init_callback=self._handle_files,
            info_callback=self._handle_info_update,
            error_callback=self._handle_error,
            result_callback=self._handle_torrent_created,
        )
        self._torrent_process.start()

    async def download_torrent(self, url):
        """
        Download torrent file and replace our own torrent

        Some trackers make unpredictable modifications to our torrent file, changing the info hash
        and forcing us to download their version. In that case, this method can be called to
        overwrite our torrent file with the altered version from the tracker.

        .. important:: This method must only be called AFTER this job is finished.

        :raise errors.RequestError: if downloading or writing file fails
        """
        if not self._torrent_process:
            raise RuntimeError('Torrent creation process has not even been started yet')
        elif not self._torrent_process.is_finished:
            raise RuntimeError('Torrent creation process is not finished yet')
        else:
            _log.debug('Downloading %r to %r', url, self._torrent_path)
            response = await utils.http.get(url)
            try:
                with open(self._torrent_path, 'wb') as f:
                    f.write(response.bytes)
            except OSError as e:
                msg = str(e.strerror) if e.strerror else str(e)
                raise errors.RequestError(msg) from None

    def terminate(self, reason=None):
        """Terminate torrent creation subprocess and all tasks"""
        if self._torrent_process:
            self._torrent_process.stop()
        super().terminate(reason=reason)

    def cancel_search(self):
        """
        Stop searching for a torrent we can copy hashes from

        If we are not currently searching for a torrent, do nothing.
        """
        if self._activity == 'searching' and self._torrent_process:
            self._torrent_process.send(utils.daemon.MsgType.terminate, utils.torrent.SKIP_SEARCHING)

    @base.unless_job_is_finished
    def _handle_files(self, files):
        self.signal.emit('file_tree', files.tree)
        self.signal.emit('file_list', files.list)

    @base.unless_job_is_finished
    def _handle_info_update(self, info):
        if isinstance(info, utils.torrent.CreateTorrentProgress):
            self._activity = 'hashing'
            self.signal.emit('progress_update', info)

        elif isinstance(info, utils.torrent.FindTorrentProgress):
            if info.status == 'verifying':
                self._activity = 'verifying'
            elif info.status == 'hit':
                _log.debug('Reusing torrent: %r', info.filepath)
                self._activity = ''
            else:
                self._activity = 'searching'
            self.signal.emit('progress_update', info)

        else:
            self._activity = ''
            raise RuntimeError('Unexpected info update: {info!r}')

    @base.unless_job_is_finished
    def _handle_torrent_created(self, torrent_path=None):
        _log.debug('Torrent created: %r', torrent_path)
        if torrent_path:
            self._torrent_path = torrent_path
            self.add_output(torrent_path)

    @base.unless_job_is_finished
    def _handle_error(self, error):
        if isinstance(error, (str, errors.TorrentCreateError)):
            self.error(error)
        else:
            self.exception(error)


def _torrent_process(output_queue, input_queue, *args, **kwargs):
    def init_callback(files):
        typ, _msg = utils.daemon.read_input_queue(input_queue)
        if typ == utils.daemon.MsgType.terminate:
            return 'cancel'
        else:
            output_queue.put((utils.daemon.MsgType.init, files))

    def progress_callback(progress):
        typ, msg = utils.daemon.read_input_queue(input_queue)
        if typ == utils.daemon.MsgType.terminate:
            if (
                    # We were told to skip searching (not to cancel torrent creation altogether)
                    msg == utils.torrent.SKIP_SEARCHING
                    # We are searching
                    and isinstance(progress, utils.torrent.FindTorrentProgress)
                    # No match found yet (False = no match, None = Verifying potential match)
                    and progress.status is not True
            ):
                # Stop searching for reusable torrent
                return utils.torrent.SKIP_SEARCHING
            else:
                # Cancel torrent creation altogether
                return 'cancel'
        else:
            output_queue.put((utils.daemon.MsgType.info, progress))

    kwargs['init_callback'] = init_callback
    kwargs['progress_callback'] = progress_callback
    try:
        torrent_path = utils.torrent.create(*args, **kwargs)
    except errors.TorrentCreateError as e:
        output_queue.put((utils.daemon.MsgType.error, e))
    else:
        output_queue.put((utils.daemon.MsgType.result, torrent_path))


class AddTorrentJob(base.JobBase):
    """
    Add torrent(s) to a BitTorrent client

    This job adds the following signals to :attr:`~.JobBase.signal`:

        ``adding``
            Emitted when attempting to add a torrent. Registered callbacks get
            the path to the torrent file as a positional argument.

        ``added``
            Emitted when the torrent was added successfully. Registered
            callbacks get the added torrent's info hash as a positional
            argument.
    """

    name = 'add-torrent'
    label = 'Add Torrent'
    cache_id = None  # Don't cache output

    @property
    def hidden(self):
        """
        `True` while this job is not :attr:`started <upsies.jobs.base.JobBase.is_started>`
        (i.e. waiting for a torrent to add), then `False`
        """
        return not self.is_started

    def initialize(self, *, btclient_config, torrent_files=(), download_path=None):
        """
        Set internal state

        :param btclient_config: :class:`~.btclient.BtclientConfig` instance
        :param torrent_files: Sequence of torrent file paths to add
        :param download_path: Download directory of each torrent's files

            This path is translated from a path on the computer that runs this code to a path on the
            BitTorrent client computer. See :attr:`~.btclients.fields.translate_path` and
            :class:`~.PathTranslations`.

        After all `torrent_files` are added, if a job named "torrent" exists,
        :meth:`~.JobBase.receive_all` torrents from that job and add them too.
        """
        self._torrent_filepaths = tuple(torrent_files)
        self._btclient_config = btclient_config
        self._download_path = download_path
        self.signal.add('adding')
        self.signal.add('added')
        self.signal.register('adding', lambda tp: setattr(self, 'info', f'Adding {utils.fs.basename(tp)}'))
        self.signal.register('added', lambda _: setattr(self, 'info', ''))
        self.signal.register('finished', lambda _: setattr(self, 'info', ''))
        self.signal.register('error', lambda _: setattr(self, 'info', ''))

        if not self._btclient_config['client']:
            self.error('No client specified')

    async def run(self):
        for torrent_filepath in self._torrent_filepaths:
            await self._add_torrent(torrent_filepath)

        if 'torrent' in self.siblings:
            async for torrent_filepath, in self.receive_all('torrent', 'output', only_posargs=True):
                await self._add_torrent(torrent_filepath)

    async def _add_torrent(self, torrent_path):
        self.signal.emit('adding', torrent_path)
        try:
            infohash = await self._btclient.add_torrent(
                torrent_path,
                download_path=self._download_path_translated,
            )
        except errors.TorrentAddError as e:
            self.error(e)
        else:
            self.add_output(infohash)
            self.signal.emit('added', infohash)

    @functools.cached_property
    def _btclient(self):
        try:
            return btclients.Btclient.from_config(self._btclient_config)
        except errors.ConfigError as e:
            # Translate ConfigError to TorrentAddError so it is caught by _add_torrent().
            raise errors.TorrentAddError(e) from e

    @functools.cached_property
    def _download_path_translated(self):
        if self._download_path:
            # Translate the (local) path that was provided to us to the (remote) path the client
            # should use to find/download the torrent's files.
            return self._btclient_config['translate_path'].translate(
                os.path.abspath(self._download_path)
            )
        else:
            # Use client's default path.
            return None


class CopyTorrentJob(base.JobBase):
    """
    Copy file(s)

    This job adds the following signals to :attr:`~.JobBase.signal`:

        ``copying``
            Emitted when attempting to copy a file. Registered callbacks get the
            source file path as a positional argument.

        ``copied``
            Emitted when the copy attempt ended. Registered callbacks get the
            destination file path (success) or the source file path (failure) as
            a positional argument.
    """

    name = 'copy-torrent'
    label = 'Copy Torrent'
    cache_id = None  # Don't cache output

    @property
    def hidden(self):
        """
        `True` while this job is not :attr:`started <upsies.jobs.base.JobBase.is_started>`
        (i.e. waiting for a torrent to add), then `False`
        """
        return not self.is_started

    def initialize(self, *, destination, torrent_files=()):
        """
        Set internal state

        :param destination: Where to put the torrent(s)
        :param torrent_files: Sequence of torrent file paths to copy to `destination`

        After all `torrent_files` are copied, if a job named "torrent" exists,
        :meth:`~.JobBase.receive_all` torrents from that job and copy them too.
        """
        self._torrent_filepaths = tuple(torrent_files)
        self._destination = None if not destination else str(destination)
        self.signal.add('copying')
        self.signal.add('copied')
        self.signal.register('copying', lambda fp: setattr(self, 'info', f'Copying {utils.fs.basename(fp)}'))
        self.signal.register('copied', lambda _: setattr(self, 'info', ''))
        self.signal.register('finished', lambda _: setattr(self, 'info', ''))
        self.signal.register('error', lambda _: setattr(self, 'info', ''))

    async def run(self):
        for torrent_filepath in self._torrent_filepaths:
            await self._copy_torrent(torrent_filepath)

        if 'torrent' in self.siblings:
            async for torrent_filepath, in self.receive_all('torrent', 'output', only_posargs=True):
                await self._copy_torrent(torrent_filepath)

    MAX_FILE_SIZE = 10 * 2**20  # 10 MiB
    """Upper limit of acceptable file size"""

    async def _copy_torrent(self, torrent_filepath):
        _log.debug('Copying %s to %s', torrent_filepath, self._destination)

        if not os.path.exists(torrent_filepath):
            self.error(f'{torrent_filepath}: No such file')

        elif os.path.getsize(torrent_filepath) > self.MAX_FILE_SIZE:
            self.error(f'{torrent_filepath}: File is too large')

        else:
            self.signal.emit('copying', torrent_filepath)

            import shutil
            try:
                new_path = shutil.copy2(torrent_filepath, self._destination)
            except OSError as e:
                msg = e.strerror if e.strerror else str(e)
                self.error(f'Failed to copy {torrent_filepath} to {self._destination}: {msg}')
            else:
                self.add_output(new_path)
                self.signal.emit('copied', new_path)



from ... import errors
from .. import utils
from ..base import rules

import logging  # isort:skip
_log = logging.getLogger(__name__)


class CbrBannedGroup(rules.BannedGroup):
    # /wikis/28
    banned_groups = {
        '3LTON',
        '4yEo',
        'ADE',
        'AFG',
        'AniHLS',
        'AnimeRG',
        'AniURL',
        'AROMA',
        'ASM',
        'aXXo',
        'BLUDV',
        'CaNNIBal',
        'CHD',
        'CM8',
        'Comando',
        'CrEwSaDe',
        'd3g',
        'DeadFish',
        'DNL',
        'DragsterPS',
        'DRENAN',
        'ELiTE',
        'eSc',
        'FaNGDiNG0',
        'FGT',
        'Flights',
        'FRDS',
        'FUM',
        'HAiKU',
        'HD2DVD',
        'HDS',
        'HDTime',
        'Hi10',
        'Hiro360',
        'ION10',
        'iPlanet',
        'JIVE',
        'KiNGDOM',
        'Lapumia',
        'Leffe',
        'LEGi0N',
        'LOAD',
        'MACCAULAY',
        'MeGusta',
        'mHD',
        'mSD',
        'NhaNc3',
        'nHD',
        'nikt0',
        'NOIVTC',
        'nSD',
        'OFT',
        'Oj',
        'PiRaTeS',
        'PlaySD',
        'playXD',
        'PRODJi',
        'RAPiDCOWS',
        'RARBG',
        'RDN',
        'REsuRRecTioN',
        'RetroPeeps',
        'RMTeam',
        'S74Ll10n',
        'SANTi',
        'SicFoI',
        'SILVEIRATeam',
        'SPASM',
        'SPDVD',
        'STUTTERSHIT',
        'Telly',
        'TGx',
        'TM',
        'TRiToN',
        'UPiNSMOKE',
        'URANiME',
        'WAF',
        'x0r',
        'xRed',
        'XS',
        'YIFY',
        'ZKBL',
        'ZmN',
        'ZMNT',
    }


class CbrPortuguese(rules.TrackerRuleBase):
    """
    Check if has Portuguese audio or subtitles
    """

    async def _check(self):
        audio_languages = utils.mediainfo.audio.get_audio_languages(
            self.release_name.path,
            exclude_commentary=True,
        )
        subtitle_languages = utils.mediainfo.text.get_subtitles(self.release_name.path)

        _log.debug('Audio languages: %r', audio_languages)
        _log.debug('Subtitle languages: %r', subtitle_languages)

        valid_dub = any(
            language in ('pt', 'zx', 'un')
            for language in audio_languages
        )
        valid_sub = any(
            subtitle.language == 'pt'
            for subtitle in subtitle_languages
        )

        if not valid_dub and not valid_sub:
            raise errors.RuleBroken('Portuguese audio or subtitles is required')



"""
Concrete :class:`~.TrackerConfigBase` subclass for NBL
"""

import base64

from ... import utils
from .. import base

import logging  # isort:skip
_log = logging.getLogger(__name__)


class NblTrackerConfig(base.TrackerConfigBase):
    upload_url: base.config.upload_url(
        base64.b64decode('aHR0cHM6Ly9uZWJ1bGFuY2UuaW8vdXBsb2FkLnBocA==').decode('ascii'),
    )

    apikey: base.config.apikey(
        default='',
        instructions='Create one on the website: <USERNAME> -> Settings -> API keys (check "Upload")',
    )

    announce_url: base.config.announce_url(
        default='',
        instructions='Get it from the website: Shows -> Upload -> Your personal announce URL',
    )

    exclude: base.config.exclude(
        base.exclude_regexes.garbage,
    )

    anonymous: base.config.anonymous('no')


cli_arguments = {
    'submit': {
        ('--anonymous', '--an'): {
            'help': (
                'Hide your username for this submission\n'
                'Valid BOOL values: ' + ', '.join(
                    f'{t}/{f}'
                    for t, f in zip(utils.types.Bool.truthy, utils.types.Bool.falsy)
                )
            ),
            'type': utils.argtypes.bool_or_none,
            'metavar': 'BOOL',
        },
        ('--tvmaze', '--tv'): {
            'help': 'TVmaze ID or URL',
            'type': utils.argtypes.webdb_id('tvmaze'),
        },
    },
}



"""
Concrete :class:`~.TrackerBase` subclass for CBR
"""

import functools
import re
import urllib

from ... import errors, utils
from ..base import TrackerBase
from . import config, rules
from .jobs import CbrTrackerJobs

import logging  # isort:skip
_log = logging.getLogger(__name__)


class CbrTracker(TrackerBase):
    name = 'cbr'
    label = 'CBR'
    torrent_source_field = 'CapybaraBR'

    setup_howto_template = (
        '{howto.introduction}\n'
        '\n'
        '{howto.next_section}. API Key\n'
        '\n'
        '   {howto.current_section}.1 On the website, go to your profile > Settings > API Key.\n'
        '   {howto.current_section}.2 If you have no API key yet, click on "Generate API key".\n'
        '   {howto.current_section}.3 Copy your personal API_KEY.\n'
        '   {howto.current_section}.4 $ upsies set trackers.{tracker.name}.apikey API_KEY\n'
        '\n'
        '{howto.next_section}. Announce URL\n'
        '\n'
        '   {howto.current_section}.1 On the website: Torrents > Upload > URL de Anúncio\n'
        '   {howto.current_section}.2 $ upsies set trackers.{tracker.name}.announce_url URL\n'
        '\n'
        '{howto.screenshots}\n'
        '\n'
        '{howto.autoseed}\n'
        '\n'
        '{howto.reuse_torrents}\n'
        '\n'
        '{howto.upload}\n'
    )

    TrackerJobs = CbrTrackerJobs
    TrackerConfig = config.CbrTrackerConfig
    cli_arguments = config.cli_arguments
    rules = rules

    async def _login(self, *, tfa_otp=None):
        pass

    async def confirm_logged_in(self):
        pass

    async def _logout(self):
        pass

    @functools.cached_property
    def _upload_url(self):
        return urllib.parse.urljoin(self.options['base_url'], '/api/torrents/upload')

    async def get_announce_url(self):
        announce_url = self.options.get('announce_url')
        if announce_url:
            return announce_url.get_secret_value()
        else:
            raise errors.AnnounceUrlNotSetError(tracker=self)

    async def upload(self, tracker_jobs):
        if not self.options['apikey']:
            raise errors.RequestError(f'trackers.{self.name}.apikey is not set')

        _log.debug('Uploading to %r', self._upload_url)
        _log.debug('POSTing data:')
        post_data = tracker_jobs.post_data
        for k, v in post_data.items():
            _log.debug(' * %s = %s', k, v)

        files = tracker_jobs.post_files
        _log.debug('Files: %r', files)

        try:
            response = await utils.http.post(
                url=self._upload_url,
                cache=False,
                user_agent=True,
                files=files,
                data=post_data,
            )
        except errors.RequestError as e:
            _log.debug(f'Request failed: {e!r}')
            _log.debug(f'url={e.url!r}')
            _log.debug(f'text={e.text!r}')
            _log.debug(f'headers={e.headers!r}')
            _log.debug(f'status_code={e.status_code!r}')
            # The error message in the HTTP response is JSON. Try to parse that
            # to get the actual error message. If that fails, raise the
            # RequestError as is.
            json = e.json(default=False)
            if not json:
                raise errors.RequestError(f'Upload failed: {e.text}') from e
        else:
            _log.debug('Upload response: %r', response)
            json = response.json()

        if json.get('success') is True:
            torrent_url = json['data']
            return self._website_url_from_torrent_url(torrent_url)
        elif json.get('message') and json.get('data'):
            raise errors.RequestError(f'Upload failed: {json["message"]}\n{json["data"]}')
        elif json.get('message'):
            raise errors.RequestError(f'Upload failed: {json["message"]}')
        else:
            raise RuntimeError(f'Unexpected response: {json!r}')

    def _website_url_from_torrent_url(self, torrent_url):
        # Torrent URL: .../torrent/download/<torrent id>.<some kind of hash?>
        # Website URL: .../torrents/<torrent id>
        website_url = torrent_url.replace('/torrent/download/', '/torrents/')
        website_url = re.sub(r'\.[a-zA-Z0-9]+$', '', website_url)
        return website_url



"""
Dummy tracker for testing and debugging
"""

import asyncio
import functools
import os
import random
import string

from .. import errors, jobs, uis, utils
from . import base

import logging  # isort:skip
_log = logging.getLogger(__name__)


if utils.is_running_in_development_environment():

    DummyImagehost = utils.types.Imagehost(allowed=('dummy',))


    class DummyTrackerConfig(base.TrackerConfigBase):
        base_url: base.config.base_url('http://localhost')
        username: base.config.username('')
        password: base.config.password('')
        tfa: utils.config.fields.boolean(
            default='no',
            description='Whether to ask for 2FA one-time password for each login.',
        )

        image_host: base.config.image_host(
            DummyImagehost,
            default=('dummy',),
        )

        exclude: base.config.exclude()

        source: utils.config.fields.string(
            default='DMY',
            description='"source" field in generated torrents.',
        )


    cli_arguments = {
        'submit': {
            ('--imdb', '--im'): {
                'help': 'IMDb ID or URL',
                'type': utils.argtypes.webdb_id('imdb'),
            },
            ('--tmdb', '--tm'): {
                'help': 'TMDb ID or URL',
                'type': utils.argtypes.webdb_id('tmdb'),
            },
            ('--skip-category', '-C'): {
                'help': 'Do not ask for category',
                'action': 'store_true',
            },
            ('--screenshots-count', '--ssc'): {
                'help': 'How many screenshots to make',
                'type': utils.argtypes.make_integer(min=3, max=10),
            },
            ('--screenshots', '--ss'): {
                'help': (
                    'Path(s) to created screenshot file(s)\n'
                    'Directories are searched recursively.\n'
                    'More screenshots are created if necessary.'
                ),
                'nargs': '+',
                'action': 'extend',
                'type': utils.argtypes.files_with_extension('png'),
            },
            ('--delay', '-d'): {
                'help': 'Number of seconds login, upload and logout take each',
                'type': float,
                'default': 1.0,
            },
            ('--get-announce-from-website'): {
                'help': 'Whether to act like we are getting the announce URL from the website ',
                'action': 'store_true',
            },
            ('--get-torrent-from-website'): {
                'help': 'Whether to download the torrent from the website instead of adding our own ',
                'action': 'store_true',
            },
            ('--poster'): {
                'help': 'Poster file or URL ',
            },
        },
        'torrent-create': {
            ('--delay', '-d'): {
                'help': 'Number of seconds login and logout take each',
                'type': float,
                'default': 0.0,
            },
            ('--get-announce-from-website'): {
                'help': 'Whether to act like we are getting the announce URL from the website ',
                'action': 'store_true',
            },
        },
    }


    class DummyTrackerJobs(base.TrackerJobsBase):

        release_name_separator = '_'

        @functools.cached_property
        def jobs_before_upload(self):
            return (
                self.login_job,

                # Interactive jobs
                self.playlists_job,
                self.tmdb_job,
                self.imdb_job,
                self.release_name_job,
                self.category_job,
                self.scene_check_job,

                # Background jobs
                self.create_torrent_job,
                self.screenshots_job,
                self.upload_screenshots_job,
                self.poster_job,
                self.mediainfo_job,

                # Silly jobs
                self.random_string_job,
                self.stupid_question_job,
                self.say_something_job,
                self.rules_job,
            )

        @functools.cached_property
        def random_string_job(self):
            return jobs.custom.CustomJob(
                name=self.get_job_name('random-string'),
                label='Random String',
                precondition=self.make_precondition('random_string_job'),
                worker=self.generate_random_string,
                catch=(
                    errors.RequestError,
                ),
                **self.common_job_args(ignore_cache=True),
            )

        async def generate_random_string(self, job):
            # await asyncio.sleep(3)
            # self.random_string_job.warn('Watch out!')

            await asyncio.sleep(1)

            # raise errors.RequestError('foo :(')
            # raise RuntimeError('foo D:')

            return ''.join(
                random.choice(string.ascii_lowercase)
                for _ in range(30)
            )

        @functools.cached_property
        def stupid_question_job(self):
            return jobs.dialog.ChoiceJob(
                name=self.get_job_name('stupid-question'),
                label='Stupid Question',
                precondition=self.make_precondition('stupid_question_job'),
                prejobs=(
                    self.random_string_job,
                ),
                autodetect=self.autodetect_stupid_answer,
                question='Is the random string random enough?',
                options=(
                    'Possibly',
                    'Maybe',
                    'Perhaps',
                ),
                multichoice=True,
                validate=self._validate_chosen,
                callbacks={
                    'finished': lambda job: job.clear_warnings(),
                },
                **self.common_job_args(),
            )

        def _validate_chosen(self, chosen):
            labels = [option[0] for option in chosen]
            if 'Perhaps' in labels:
                raise ValueError('I have decided that "Perhaps" is not an option after all.')
            elif len(labels) < 2:
                raise ValueError('Pick at least 2 options.')

        async def autodetect_stupid_answer(self, job):
            assert self.random_string_job.is_finished

            # raise RuntimeError('How would I know?')

            random_string = self.random_string_job.output[0]
            if all(character in random_string for character in 'abc'):
                return 'Maybe'
            else:
                self.stupid_question_job.add_prompt(
                    uis.prompts.RadioListPrompt(
                        question='Is this job too stupid?',
                        options=('Yes', 'No'),
                        callbacks=(
                            self.prompt_stupid_answer_callback,
                        ),
                    )
                )
                self.stupid_question_job.add_prompt(
                    uis.prompts.CheckListPrompt(
                        question='Are you sure?',
                        options=('Yes', 'No', 'Maybe?'),
                        callbacks=(
                            self.prompt_stupid_answer_callback_2,
                        ),
                    )
                )
                self.stupid_question_job.add_prompt(
                    uis.prompts.RadioListPrompt(
                        options=('Foo', 'Bar'),
                        callbacks=(),
                    )
                )

        def prompt_stupid_answer_callback(self, result):
            if result == 'Yes':
                self.stupid_question_job.error('This job is too stupid.')
            else:
                self.stupid_question_job.warn('Choose wisely!')

        def prompt_stupid_answer_callback_2(self, result):
            if 'Yes' in result and 'No' in result:
                self.stupid_question_job.error('Yes and No? Be more decisive!')
            elif 'Maybe?' in result:
                self.stupid_question_job.warn('Okay?')
            elif 'Yes' not in result:
                self.stupid_question_job.error('This job may be too stupid after all!')

        @functools.cached_property
        def say_something_job(self):
            return jobs.dialog.TextFieldJob(
                name=self.get_job_name('say-something'),
                label='Something',
                precondition=self.make_precondition('say_something_job'),
                prejobs=(
                    self.random_string_job,
                ),
                warn_exceptions=(
                    ValueError,
                ),
                text=self.generate_something,
                **self.common_job_args(),
            )

        async def generate_something(self):
            assert self.random_string_job.is_finished
            word = self.random_string_job.output[0]

            await asyncio.sleep(6)
            self.say_something_job.warn('Still thinking...')
            await asyncio.sleep(6)

            self.say_something_job.clear_warnings()
            if 'x' in word:
                raise ValueError(f'There is an "x" in "{word}"! Yuck!')
            elif 'a' in word:
                return 'Avocado!'
            else:
                return f'This is a completely original random string: {word}'

        @functools.cached_property
        def category_job(self):
            if not self.options['skip_category']:
                return jobs.dialog.ChoiceJob(
                    name=self.get_job_name('category'),
                    label='Category',
                    precondition=self.make_precondition('category_job'),
                    options=(
                        (str(typ).capitalize(), typ)
                        for typ in utils.types.ReleaseType if typ
                    ),
                    autodetected=self.release_name.type,
                    **self.common_job_args(),
                )


    class DummyHdOnly(base.rules.HdOnly):
        message = 'This is not HD!'

        async def _check(self):
            await super()._check()


    class DummyBannedGroup(base.rules.BannedGroup):
        banned_groups = {
            'FoO',
            'BAR',
        }

        async def _check_custom(self):
            if (
                    self.is_group('BAZ')
                    and 'Remux' in self.release_name.source
            ):
                raise errors.BannedGroup('BAZ', additional_info='No remuxes')


    class DummyNoEpisodes(base.rules.TrackerRuleBase):
        required_jobs = ('category_job', 'say_something_job', 'imdb_job')

        async def _check(self):
            _log.debug('################# CHECKING CATEGORY: %r', self.tracker_jobs.category_job.choice)
            await super()._check()
            if self.tracker_jobs.category_job.choice is utils.types.ReleaseType.episode:
                raise errors.RuleBroken('No episodes allowed')


    class DummyTracker(base.TrackerBase):
        name = 'dummy'
        label = 'DuMmY'

        cli_arguments = cli_arguments
        rules = (
            DummyBannedGroup,
            DummyHdOnly,
            DummyNoEpisodes,
        )

        @property
        def torrent_source_field(self):
            return self.options['source']

        setup_howto_template = (
            'This is just a no-op tracker for testing and demonstration.'
        )

        TrackerJobs = DummyTrackerJobs
        TrackerConfig = DummyTrackerConfig

        async def _login(self, tfa_otp=None):
            username, password = self.options['username'], self.options['password']
            _log.debug('%s: Logging in as %s (password=%s, OTP=%s)', self.name, username, password, tfa_otp)
            if self.options['tfa'] and not tfa_otp:
                raise errors.TfaRequired('I need a a second factor.')

            await asyncio.sleep(self.options['delay'])

            # Simulate wrong OTP.
            if tfa_otp and not all(c in '1234567890' for c in tfa_otp):
                raise errors.RequestError('Your OTP looks weird.')

            # Simulate wrong username/password combination.
            if self.options['password'] == self.options['username']:
                raise errors.RequestError('Your username and password are identical, dummy!')

            _log.debug('%s: Logged in as %s (password=%s, OTP=%s)', self.name, username, password, tfa_otp)

        async def confirm_logged_in(self):
            _log.debug('%s: Confirming logged in', self.name)

        async def _logout(self):
            _log.debug('%s: Logging out', self.name)
            await asyncio.sleep(self.options['delay'])
            # raise errors.RequestError('You shall not logout!')
            _log.debug('%s: Logged out', self.name)

        async def get_announce_url(self):
            if self.options['get_announce_from_website']:
                _log.debug('%s: Getting announce URL from website', self.name)
                await asyncio.sleep(1)
            else:
                _log.debug('%s: Getting announce URL from config file', self.name)
            return 'http://localhost:123/f1dd15718/announce'

        async def upload(self, tracker_jobs):
            if tracker_jobs.create_torrent_job.output:
                torrent_file = tracker_jobs.create_torrent_job.output[0]
            else:
                raise errors.RequestError('Torrent file was not created.')
            _log.debug('%s: Uploading %s', self.name, torrent_file)
            await asyncio.sleep(self.options['delay'])

            if self.options['get_torrent_from_website']:
                torrent_url = 'http://localhost:1234/my_altered_torrent.torrent'
                _log.debug('%s: Downloading %s', self.name, torrent_url)
                await tracker_jobs.create_torrent_job.download_torrent(torrent_url)

            return f'http://localhost/{os.path.basename(torrent_file)}'



"""
Concrete :class:`~.TrackerJobsBase` subclass for FLD
"""

import functools
import re

from ... import jobs, utils
from ..base import TrackerJobsBase

import logging  # isort:skip
_log = logging.getLogger(__name__)


class FldTrackerJobs(TrackerJobsBase):

    @functools.cached_property
    def jobs_before_upload(self):
        # NOTE: Keep in mind that the order of jobs is important for
        #       isolated_jobs: The final job is the overall result, so if
        #       upload_screenshots_job is listed after description_job,
        #       --only-description is going to print the list of uploaded
        #       screenshot URLs.
        return (
            # Interactive jobs
            self.playlists_job,
            self.tmdb_job,
            self.imdb_job,
            self.release_name_job,
            self.category_job,

            # Background jobs
            self.create_torrent_job,
            self.mediainfo_job,
            self.bdinfo_job,
            self.screenshots_job,
            self.upload_screenshots_job,
            self.description_job,
            self.rules_job,
        )

    @property
    def isolated_jobs(self):
        """
        Sequence of job attribute names (e.g. "imdb_job") that were singled
        out by the user, e.g. with a CLI argument
        """
        if self.options.get('only_description', False):
            return self.get_job_and_dependencies(self.description_job)
        else:
            # Activate all jobs in jobs_before/after_upload
            return ()

    @functools.cached_property
    def tmdb_job(self):
        tmdb_job = super().tmdb_job
        tmdb_job.no_id_ok = True
        return tmdb_job

    @functools.cached_property
    def category_job(self):
        return jobs.dialog.ChoiceJob(
            name=self.get_job_name('category'),
            label='Category',
            precondition=self.make_precondition('category_job'),
            prejobs=(
                self.release_name_job,
            ),
            autodetect=self.autodetect_category,
            autofinish=True,
            options=(
                ('Movie', 'movie'),
                ('TV Season', 'show_season'),
                ('TV Episode', 'show_episode'),
            ),
            **self.common_job_args(),
        )

    _autodetect_category_map = {
        'Movie': lambda release_name: release_name.type is utils.release.ReleaseType.movie,
        'TV Season': lambda release_name: release_name.type is utils.release.ReleaseType.season,
        'TV Episode': lambda release_name: release_name.type is utils.release.ReleaseType.episode,
    }

    def autodetect_category(self, _):
        approved_release_name = self.release_name
        _log.debug('Autodetecting category: Approved release type: %r', approved_release_name.type)
        for label, is_match in self._autodetect_category_map.items():
            if is_match(approved_release_name):
                return label

    release_name_english_title_before_original = True
    release_name_translation = {
        'source': {
            re.compile(r'(?i:remux)'): 'REMUX',
            re.compile(r'(?i:hybrid)'): 'HYBRID',
        },
    }

    @functools.cached_property
    def description_job(self):
        return jobs.dialog.TextFieldJob(
            name=self.get_job_name('description'),
            label='Description',
            precondition=self.make_precondition('description_job'),
            prejobs=(
                self.playlists_job,
                self.mediainfo_job,
                self.bdinfo_job,
                self.screenshots_job,
                self.upload_screenshots_job,
            ),
            text=self.generate_description,
            hidden=True,
            finish_on_success=True,
            read_only=True,
            **self.common_job_args(ignore_cache=True),
        )

    image_host_config = {
        'common': {'thumb_width': 350},
    }

    def generate_description(self):
        screenshots = self._generate_description_screenshots()
        mediainfos = self._generate_description_mediainfos()

        description_parts = [
            '[center]\n'
            + screenshots
            + '\n[/center]'
        ]

        if mediainfos:
            # Include mediainfos if there are multiple (e.g. for IFO/VOB or if `document_all_videos`
            # is True)
            description_parts.append(
                '\n[center][h3]Mediainfo[/h3][/center]\n'
                + mediainfos
            )

        if promo := self.generate_promotion_bbcode():
            description_parts.append(f'\n\n\n{promo}')

        return ''.join(description_parts)

    def _generate_description_screenshots(self):
        assert self.upload_screenshots_job.is_finished
        return utils.bbcode.screenshots_grid(
            screenshots=self.upload_screenshots_job.uploaded_images,
            columns=2,
            horizontal_spacer='   ',
            vertical_spacer='\n\n',
        )

    def _generate_description_mediainfos(self):
        mediainfo_tags = []
        for video_filepath, mediainfo in self._description_mediainfos.items():
            filetitle = self.get_relative_file_path(video_filepath)
            mediainfo_tags.append(
                f'[hide={filetitle}][code]{mediainfo}[/code][/hide]'
            )
        return '\n'.join(mediainfo_tags)

    # Only generate mediainfo/bdinfo reports and screenshots for the first/main video file/playlist.
    document_all_videos = False

    @property
    def _description_mediainfos(self):
        # mediainfo_job should be disabled for BDMV releases. In that case, the BDInfo report is
        # passed normally in place of the mediainfo report. (See mediainfo_filehandle.)
        if self.mediainfo_job.is_enabled:
            assert self.mediainfo_job.is_finished

            # For VIDEO_TS releases, there should be one mediainfo report for a .IFO and another for
            # a .VOB file. The .IFO report should always be passed separately, but we include them
            # both in the description for ease of access.
            mediainfos_by_file = self.mediainfo_job.reports_by_file
            if len(mediainfos_by_file) >= 2:
                return mediainfos_by_file

        return {}

    @property
    def post_data(self):
        return {
            'name': self.get_job_output(self.release_name_job, slice=0),
            'media_type': self.get_job_attribute(self.category_job, 'choice'),
            'media_info': self.get_job_output(self.mediainfo_job, slice=0),
            'imdb_id': self.get_job_output(self.imdb_job, slice=0),
            'tmdb_id': self.get_job_output(self.tmdb_job, slice=0),
            'description': self.get_job_output(self.description_job, slice=0),
            'anonymous': self.options['anonymous'],
        }



import os
import re

# Regular expressions are matched against file paths.

# Path separator must be escaped in case we're running on Windows.
sep = re.escape(os.sep)

checksums = r'\.(?i:sfv|md5)$'

extras = (
    r'(?i:'
    # Directory named "extras".
    rf'{sep}extras{sep}'
    r'|'
    # Anything with the word "extras" with another word in front of it. This
    # should exclude the show "Extras".
    rf'{sep}.+[\. ]extras[\. ]'
    r'|'
    # Numbered extras (e.g. "Foo.S01.Extra1.mkv", "Foo.S01.Extra.2.mkv", .etc)
    rf'{sep}.+[\. ]extra[\. ]?\d+[\.]'
    r')'
)

images = (
    r'(?i:'
    # Directory is not a DISC
    rf'^(?!.*{sep}(BDMV|VIDEO_TS){sep}).*'
    # Ends with image types
    r'\.(png|jpg|jpeg)$'
    r')'
)

nfo = r'\.(?i:nfo)$'

samples = (
    r'(?i:'
    # Sample directory
    rf'{sep}[!_0-]?sample{sep}'
    r'|'
    # Sample file name starts with release name
    rf'[^{sep}][\.\-_ ]sample\.mkv'
    r'|'
    # Sample file name ends with release name
    rf'{sep}sample[\!\-_].+\.mkv'
    r'|'
    # Sample file name starts with release name and ends with "sample-RLSGRP.mkv"
    r'[\.\-_!]?sample-[a-zA-Z0-9]+\.mkv'
    r'|'
    # Sample file name starts with "<characters that top-sort>sample"
    rf'{sep}[!#$%&*+\-\.]?sample\.mkv'
    r')'
)

subtitles = r'\.(?i:srt|idx|sub)$'

garbage = (
    f'{sep}(?:'
    + '|'.join((
        # Linux
        r'\.fuse_hidden\d+',     # Opened + removed file on FUSE
        rf'\.Trash{sep}.*',      # Deleted files
        rf'\.Trash-\d+{sep}.*',  # Deleted files

        # MacOS
        r'\.DS_Store',         # Folder options
        rf'\.Trashes{sep}.*',  # Deleted files

        # Windows
        r'desktop\.ini',  # Folder options
        r'Thumbs\.db',    # Thumbnails

        # Synology
        rf'@eaDir{sep}.*',
        r'.*@SynoEAStream',
        r'.*@SynoResource',

        # Other software
        r'.*\.miniso',    # DVDFab
        r'.*dvdid\.xml',  # AnyDVD

        # Usenet
        r'.*\.nzb',
        r'.*\.par2',

        # srrDB
        r'.*\.srr',
        r'.*\.srs',

        # Miscellaneous
        r'.*\.torrent',
    ))
    + ')$'
)



"""
Concrete :class:`~.TrackerBase` subclass for RFX
"""

import re

from ... import errors, utils
from ..base import TrackerBase
from . import config
from .jobs import RfxTrackerJobs

import logging  # isort:skip
_log = logging.getLogger(__name__)


class RfxTracker(TrackerBase):
    name = 'rfx'
    label = 'RFX'
    torrent_source_field = 'ReelFliX'

    setup_howto_template = (
        '{howto.introduction}\n'
        '\n'
        '{howto.next_section}. API Key\n'
        '\n'
        '   {howto.current_section}.1 On the website, go to your profile > Settings > API Key.\n'
        '   {howto.current_section}.2 If you have no API_KEY yet, generate one.\n'
        '   {howto.current_section}.3 Copy your personal API_KEY.\n'
        '   {howto.current_section}.4 $ upsies set trackers.{tracker.name}.apikey API_KEY\n'
        '\n'
        '{howto.next_section}. Announce URL\n'
        '\n'
        '   {howto.current_section}.1 On the website, go to your profile > Settings > Passkey.\n'
        '   {howto.current_section}.2 Copy your personal PASSKEY.\n'
        '   {howto.current_section}.2 $ upsies set trackers.{tracker.name}.announce_passkey PASSKEY\n'
        '\n'
        '{howto.screenshots}\n'
        '\n'
        '{howto.autoseed}\n'
        '\n'
        '{howto.reuse_torrents}\n'
        '\n'
        '{howto.upload}\n'
    )

    TrackerJobs = RfxTrackerJobs
    TrackerConfig = config.RfxTrackerConfig
    cli_arguments = config.cli_arguments

    async def _login(self, *, tfa_otp=None):
        pass

    async def confirm_logged_in(self):
        pass

    async def _logout(self):
        pass

    async def get_announce_url(self):
        if not self.options['announce_passkey']:
            raise errors.RequestError(f'trackers.{self.name}.announce_passkey is not set')
        else:
            return '/'.join((
                self.options['announce_url'].rstrip('/'),
                self.options['announce_passkey'].get_secret_value(),
            ))

    async def upload(self, tracker_jobs):
        if not self.options['apikey']:
            raise errors.RequestError(f'trackers.{self.name}.apikey is not set')

        _log.debug('Uploading to %r', self.options['upload_url'])
        _log.debug('POSTing data:')
        for k, v in tracker_jobs.post_data.items():
            _log.debug(' * %s = %s', k, v)
        _log.debug('POSTing files:')
        for k, v in tracker_jobs.post_files.items():
            _log.debug(' * %s = %s', k, v)

        response = await utils.http.post(
            url=self.options['upload_url'],
            cache=False,
            user_agent=True,
            headers={'Authorization': f'Bearer {self.options["apikey"].get_secret_value()}'},
            files=tracker_jobs.post_files,
            data=tracker_jobs.post_data,
        )
        _log.debug('Upload response: %r', response)
        json = response.json()
        try:
            if json['success']:
                torrent_url = json['data']
                await tracker_jobs.create_torrent_job.download_torrent(torrent_url)
                return self._torrent_page_url_from_download_url(torrent_url)
            else:
                raise errors.RequestError(f'Upload failed: {json["message"]}')
        except KeyError as e:
            raise RuntimeError(f'Unexpected response: {response}') from e

    def _torrent_page_url_from_download_url(self, torrent_download_url):
        # Torrent URL: .../torrent/download/TORRENT_ID.RANDOM_ASCII_CHARACTERS(?)
        # Website URL: .../torrents/TORRENT_ID
        torrent_page_url = torrent_download_url.replace('/torrent/download/', '/torrents/')
        torrent_page_url = re.sub(r'\.[a-zA-Z0-9]+$', '', torrent_page_url)
        return torrent_page_url



"""
Helper functions for generating BBcode
"""

from .. import utils


def screenshots_grid(screenshots, columns=2, horizontal_spacer='  ', vertical_spacer='\n'):
    """
    Return BBcode for thumbnailed screenshots in a grid layout

    :param screenshots: Sequence of :class:`~.imagehosts.common.UploadedImage` objects (URLs with a
        ``thumbnail_url`` attribute)

    :param int columns: How many columns to split screenshots into

        `columns` may also be a sequence of acceptable values to automatically find the best number
        of columns for the number of screenshots.

        For example, for 9 screenshots and columns=(2, 3, 4), the best number of columns would be 3,
        for 8 screenshots it would be 4, and for 2 screenshots it would be 2 columns.

    :param str horizontal_spacer: String between columns
    :param str vertical_spacer: String between rows

    :raise RuntimeError: if any screenshot doesn't have a thumbnail
    """
    if isinstance(columns, int):
        columns = (columns,)

    groups = utils.as_groups(
        screenshots,
        group_sizes=columns,
        default=None,
    )

    rows = []
    for screenshots in groups:
        cells = []
        for screenshot in screenshots:
            # `screenshot` is `None` at the end if the number of screenshots is not perfectly
            # divisible by `columns`.
            if screenshot is not None:
                if screenshot.thumbnail_url is None:
                    raise RuntimeError(f'No thumbnail for {screenshot}')
                cells.append(f'[url={screenshot}][img]{screenshot.thumbnail_url}[/img][/url]')

        # Join cells horizontally.
        rows.append(horizontal_spacer.join(cells))

    # Join rows vertically.
    return vertical_spacer.join(rows)



"""
Base class for image uploaders
"""

import abc
import glob
import hashlib
import os

from .. import __project_name__, constants, errors, utils
from . import common

import logging  # isort:skip
_log = logging.getLogger(__name__)


class ImagehostConfigBase(utils.config.SubsectionBase):
    """
    Base class for image host configurations

    User configuration for an image host, for example for API keys.
    """

    thumb_width: utils.config.fields.integer(
        default=0,
        min=0,
        description=(
            'Thumbnail width in pixels or 0 for no thumbnail.\n'
            'Trackers may ignore this option and use a hardcoded thumbnail width.'
        ),
    )


class ImagehostBase(abc.ABC):
    """
    Base class for image uploaders

    :param config: Instance of :attr:`~.Config` or dictionary with keyword argument for that class
    :param str cache_directory: Where to cache URLs; defaults to
        :attr:`~upsies.constants.DEFAULT_CACHE_DIRECTORY`
    """

    supported_mime_types = ('image/png', 'image/jpeg')
    """Sequence of MIME types that are supported by the hosting service"""

    @property
    @abc.abstractmethod
    def name(self):
        """Name of the image hosting service"""

    @property
    @abc.abstractmethod
    def Config(self):
        """Subclass of :class:`~.ImagehostConfigBase`"""

    def __init__(self, config, cache_directory=None):
        if isinstance(config, self.Config):
            self._config = config
        else:
            self._config = self.Config(**config)
        self._cache_directory = cache_directory if cache_directory else constants.DEFAULT_CACHE_DIRECTORY

    @property
    def cache_directory(self):
        """Path to directory where upload info is cached"""
        return self._cache_directory

    @cache_directory.setter
    def cache_directory(self, directory):
        self._cache_directory = directory

    @property
    def config(self):
        """:class:`~.ImagehostConfigBase` instance provided during instantiation"""
        return self._config

    cli_arguments = {}
    """CLI argument definitions (see :attr:`.CommandBase.cli_arguments`)"""

    description = ''
    """Any documentation, for example how to get an API key"""

    async def upload(self, image_path, *, thumb_width=None, cache=True):
        """
        Upload image file

        :param image_path: Path to image file
        :param int thumb_width: Override ``thumb_width`` in :attr:`config`

            If set to 0, no thumbnail is uploaded.
        :param bool cache: Whether to attempt to get the image URL from cache

        :raise RequestError: if the upload fails

        :return: :class:`~.imagehost.common.UploadedImage`
        """
        if 'apikey' in self.config and not self.config['apikey']:
            raise errors.RequestError(
                'You must configure an API key first. Run '
                f'"{__project_name__} upload-images {self.name} --help" '
                'for more information.'
            )

        # Upload fullsize image.
        info = {
            'url': await self._get_image_url(image_path, cache=cache),
        }

        # Get thumbnail via argument, if specified, or from config file.
        if thumb_width is None:
            thumb_width = self.config['thumb_width']
        if thumb_width:
            # Create thumbnail.
            try:
                thumbnail_path = utils.image.resize(
                    image_path,
                    width=thumb_width,
                    target_directory=self.cache_directory,
                    overwrite=not cache,
                )
            except errors.ImageResizeError as e:
                raise errors.RequestError(e) from e
            else:
                # Upload thumbnail.
                info['thumbnail_url'] = await self._get_image_url(thumbnail_path, cache=cache)

        return common.UploadedImage(**info)

    async def _get_image_url(self, image_path, *, cache=True):
        url = self._get_url_from_cache(image_path) if cache else None
        if not url:
            try:
                converted_image_path = self._get_proper_image(image_path)
            except errors.ImageConvertError as e:
                raise errors.RequestError(e) from e
            else:
                _log.debug('Uploading image: %r', converted_image_path)
                url = await self._upload_image(converted_image_path)
                _log.debug('Uploaded image: %r: %r', converted_image_path, url)
                self._store_url_to_cache(image_path, url)
        return url

    def _get_proper_image(self, image_path):
        # Convert `image_path` to png if it is in an unsupported format.
        # Raise exception if format cannot be determined.
        mime_type = utils.image.get_mime_type(image_path)
        _log.debug('MIME type of %r: %r', image_path, mime_type)
        if not mime_type:
            raise errors.ImageConvertError(f'Unknown file type: {image_path}')
        elif mime_type not in self.supported_mime_types:
            filename_suffix = f'.{self._get_cache_file_id(image_path)}.png'
            max_filename_length = 250 - len(filename_suffix)
            filename = utils.fs.basename(image_path)[-max_filename_length:] + filename_suffix
            mime_type = 'image/png'
            output_file = os.path.join(self.cache_directory, filename)
            _log.debug('Converted to %r: %r', mime_type, output_file)
            return utils.image.convert(image_path, mime_type=mime_type, output_file=output_file)
        else:
            return image_path

    @abc.abstractmethod
    async def _upload_image(self, image_path):
        """Upload `image_path` and return URL to the image file"""

    def _get_url_from_cache(self, image_path):
        cache_file_suffix = self._get_cache_file_suffix(image_path)
        cache_file_glob = os.path.join(
            glob.escape(self.cache_directory),
            f'*.{cache_file_suffix}',
        )
        matching_cache_files = glob.glob(cache_file_glob)

        if matching_cache_files:
            cache_file = matching_cache_files[0]
            try:
                with open(cache_file, 'r') as f:
                    url = f.read().strip()
            except OSError as e:
                # Unreadable cache file. We'll try to overwrite it later.
                _log.debug('Failed to read %r: %r', cache_file, e)
            else:
                _log.debug('Got URL for %r from cache: %s: %s', image_path, cache_file, url)
                return url

    def _store_url_to_cache(self, image_path, url):
        # Prepend file name to unique ID for easier debugging. Max file name length is usually 255
        # bytes.
        cache_file_suffix = self._get_cache_file_suffix(image_path)
        max_filename_length = 250 - len(cache_file_suffix)
        filename = f'{utils.fs.basename(image_path)}.{cache_file_suffix}'
        cache_file = os.path.join(self.cache_directory, filename[-max_filename_length:])

        try:
            utils.fs.mkdir(utils.fs.dirname(cache_file))
            with open(cache_file, 'w') as f:
                f.write(url)
        except (OSError, errors.ContentError) as e:
            msg = e.strerror if getattr(e, 'strerror', None) else e
            raise RuntimeError(f'Unable to write cache {cache_file}: {msg}') from e

    def _get_cache_file_suffix(self, image_path):
        unique_id = self._get_cache_file_id(image_path)
        return f'{unique_id}.{self.name}.url'

    def _get_cache_file_id(self, image_path):
        try:
            # Generate unique ID from the first 10 KiB of image data.
            with open(image_path, 'rb') as f:
                return hashlib.md5(f.read(10 * 1024)).hexdigest()
        except OSError:
            # If `image_path` is not readable, get unique ID from file path.
            return hashlib.md5(image_path.encode('utf8')).hexdigest()



"""
API for imdb.com
"""

import functools
import json
import re

from ... import utils
from ..types import ReleaseType
from . import common
from .base import WebDbApiBase

import logging  # isort:skip
_log = logging.getLogger(__name__)


GRAPHQL_QUERY_SEARCH = """
{
    advancedTitleSearch(
        first: 30,
        sort: { sortBy: POPULARITY, sortOrder: ASC }
        constraints: {%%%CONSTRAINTS%%%}
    ) {
        total
        edges {
            node {
                title {
                    id
                    titleText {
                        text
                    }
                    originalTitleText {
                        text
                    }
                    titleType {
                        id
                    }
                    releaseYear {
                        year
                    }
                    plot {
                        plotText {
                            plainText
                        }
                    }
                    countriesOfOrigin {
                        countries {
                            text
                        }
                    }
                }
            }
        }
    }
}
"""

GRAPHQL_QUERY_ID = """
query GetTitleInfo {
    title(id: "%%%ID%%%") {
        id
        titleText {
            text
            isOriginalTitle
            country {
                text
            }
        }
        originalTitleText {
            text
        }
        releaseYear {
            year
            endYear
        }
        titleType {
            id
        }
        plot {
            plotText {
                plainText
            }
        }
        ratingsSummary {
            aggregateRating
            voteCount
        }
        primaryImage {
            url
        }
        runtime {
            displayableProperty {
                value {
                    plainText
                }
            }
            seconds
        }
        titleGenres {
            genres {
                genre {
                    text
                }
            }
        }
        principalCredits {
            category {
                text
                id
            }
            credits {
                name {
                    id
                    nameText {
                        text
                    }
                }
            }
        }
        runtimes(first: 10) {
            edges {
                node {
                    id
                    seconds
                    displayableProperty {
                        value {
                            plainText
                        }
                    }
                    attributes {
                        text
                    }
                }
            }
        }
        countriesOfOrigin {
            countries {
                text
            }
        }
    }
}
"""

TYPE_MAP = {
    ReleaseType.movie: ('movie', 'short', 'tvMovie', 'video', 'tvShort'),
    ReleaseType.season: ('tvSeries', 'tvMiniSeries'),
    # Searching for single episodes is currently not supported
    ReleaseType.episode: ('tvSeries', 'tvMiniSeries'),
}

TYPE_MAP_REVERSE = {
    'movie': ReleaseType.movie,
    'short': ReleaseType.movie,
    'tvMovie': ReleaseType.movie,
    'video': ReleaseType.movie,
    'tvShort': ReleaseType.movie,
    'tvSeries': ReleaseType.season,
    'tvMiniSeries': ReleaseType.season,
    'tvEpisode': ReleaseType.episode,
}

WEBSITE_BASE = 'https://www.imdb.com'


class ImdbApi(WebDbApiBase):
    """API for imdb.com"""

    name = 'imdb'
    label = 'IMDb'

    default_config = {}

    def get_id_from_text(self, text):
        # Example: https://www.imdb.com/title/tt0048918/
        match = re.search(r'\b(tt\d+)\b', text)
        if match:
            return match.group(1)

    def sanitize_query(self, query):
        """
        Deal with IMDb-specific quirks

        - Remove ``"and"`` from :attr:`.Query.title` because IMDb doesn't find ``"Foo & Bar"`` if we
          search for ``"Foo and Bar"``. It seems to work vice versa, i.e. the query ``"Foo and
          Bar"`` finds ``"Foo & Bar"``, so we keep any ``"&"``.

        - Replace ``"dont"`` with ``"don't"``, ``"cant"`` with ``"can't"``, etc.
        """
        query = super().sanitize_query(query)
        query.title = re.sub(r'\s(?i:and)(\s)', r'\1', query.title)
        query.title = re.sub(r'\b(?i:dont)(\b)', "don't", query.title)
        query.title = re.sub(r'\b(?i:cant)(\b)', "can't", query.title)
        query.title = re.sub(r'\b(?i:wont)(\b)', "won't", query.title)
        return query

    _url_base = 'https://caching.graphql.imdb.com'

    def _get_graphql_query(self, template_name, values):
        template = globals()[f'GRAPHQL_QUERY_{template_name.upper()}']
        query = ' '.join(template.replace('\n', ' ').split())
        for k, v in values.items():
            query = query.replace(f'%%%{k}%%%', v)
        if '%%%' in query:
            raise RuntimeError(f'Unresolved template string in query: {query}')
        else:
            return json.dumps({'query': query})

    async def _get_info(self, id):
        query = self._get_graphql_query('ID', {'ID': id})
        response = (await utils.http.post(
            url=self._url_base,
            data=query,
            headers={'Content-Type': 'application/json'},
            timeout=10,
            cache=True,
            user_agent='BROWSER',
        )).json()
        return response['data']['title']

    async def search(self, query):
        _log.debug('Searching IMDb for %r', query)
        if query.id:
            title_english = await self.title_english(query.id)
            title_original = await self.title_original(query.id)
            return [_ImdbSearchResult(
                imdb_api=self,
                cast=functools.partial(self.cast, query.id),
                countries=functools.partial(self.countries, query.id),
                directors=functools.partial(self.directors, query.id),
                genres=functools.partial(self.genres, query.id),
                id=query.id,
                summary=functools.partial(self.summary, query.id),
                title=title_english or title_original,
                title_english=title_english,
                title_original=title_original,
                type=await self.type(query.id),
                url=await self.url(query.id),
                year=await self.year(query.id),
            )]

        elif not query.title:
            return []

        else:
            constraints = [f'titleTextConstraint: {{searchTerm: "{query.title}"}}']

            if query.type is not ReleaseType.unknown:
                types = TYPE_MAP[query.type]
                types_str = '[' + ', '.join(f'"{t}"' for t in types) + ']'
                constraints.append(f'titleTypeConstraint: {{anyTitleTypeIds: {types_str}}}')

            if query.year is not None:
                constraints.append(
                    f'releaseDateConstraint: {{releaseDateRange: {{start: "{query.year}-01-01", end: "{query.year}-12-31"}}}}'
                )

            constraints.append('explicitContentConstraint: {explicitContentFilter: INCLUDE_ADULT}')

            query = self._get_graphql_query('SEARCH', {'CONSTRAINTS': ', '.join(constraints)})
            response = (await utils.http.post(
                url=self._url_base,
                data=query,
                headers={'Content-Type': 'application/json'},
                timeout=10,
                cache=True,
                user_agent='BROWSER',
            )).json()
            return [
                _ImdbSearchResult(info=result['node']['title'], imdb_api=self)
                for result in response['data']['advancedTitleSearch']['edges']
            ]

    async def cast(self, id):
        if id:
            return await self._get_persons(id, category='cast')
        return ()

    async def creators(self, id):
        if id:
            info = await self._get_info(id)
            _log.debug(json.dumps(info, indent=4))
            return await self._get_persons(id, category='creator')
        return ()

    async def directors(self, id):
        if id:
            return await self._get_persons(id, category='director')
        return ()

    async def _get_persons(self, id, *, category):
        info = await self._get_info(id)
        principal_credits = info.get('principalCredits', ())
        credits = ()
        for credits_ in principal_credits:
            category_ = credits_.get('category') or {}
            if category_.get('id') == category:
                credits = credits_.get('credits', ())
                break

        def get_person(credit):
            name = ((credit.get('name') or {}).get('nameText') or {}).get('text', '')
            id = (credit.get('name') or {}).get('id', '')
            if name and id:
                return common.Person(name, url=f'{WEBSITE_BASE}/name/{id}')

        return tuple(
            person
            for credit in credits
            if (person := get_person(credit))
        )

    async def _countries(self, id):
        countries = []
        if id:
            info = await self._get_info(id)
            items = (info.get('countriesOfOrigin') or {}).get('countries', ())
            countries.extend(
                country
                for item in items
                if (country := item.get('text'))
            )
        return tuple(countries)

    async def genres(self, id):
        genres = []
        if id:
            info = await self._get_info(id)
            items = (info.get('titleGenres') or {}).get('genres', ())
            genres.extend(
                genre.casefold()
                for item in items
                if (genre := ((item.get('genre') or {}).get('text') or None))
            )
        return tuple(genres)

    async def poster_url(self, id, season=None):
        if id:
            info = await self._get_info(id)
            poster_url = (info.get('primaryImage') or {}).get('url', '')
            # Request scaled down poster (300 pixels wide)
            poster_url = re.sub(r'._V1_*.jpg$', '._V1_SX300.jpg', poster_url)
            return poster_url
        return ''

    rating_min = 0.0
    rating_max = 10.0

    async def rating(self, id):
        if id:
            info = await self._get_info(id)
            return (info.get('ratingsSummary') or {}).get('aggregateRating', None)
        return None

    _ignored_runtimes_keys = (
        re.compile(r'^(?i:approx)\w*$'),
    )

    async def _runtimes(self, id):
        if id:
            info = await self._get_info(id)
            runtimes = tuple(
                node
                for edge in (info.get('runtimes') or {}).get('edges', ())
                if (node := edge.get('node'))
            )

            def get_cut_name(runtime):
                attributes = runtime.get('attributes', ())
                name = 'default'
                if attributes:
                    name = attributes[0].get('text') or 'default'
                if name != 'default':
                    # Capitalize words. We can't use "\b" because that results in "Director'S Cut".
                    name = re.sub(r'(?:^|\s)[a-z]', lambda match: match.group(0).upper(), name)
                return name

            def get_runtime_minutes(runtime):
                return round(runtime.get('seconds', 0) / 60)

            return {
                get_cut_name(runtime): get_runtime_minutes(runtime)
                for runtime in runtimes
            }

        return {}

    async def summary(self, id):
        if id:
            info = await self._get_info(id)
            return ((info.get('plot') or {}).get('plotText') or {}).get('plainText', '')
        return ''

    async def _title_original(self, id):
        if id:
            info = await self._get_info(id)
            return (info.get('originalTitleText') or {}).get('text', '')
        return ''

    async def _titles_english(self, id):
        if id:
            info = await self._get_info(id)
            return ((info.get('titleText') or {}).get('text', ''),)
        return ()

    async def type(self, id):
        if id:
            info = await self._get_info(id)
            name = (info.get('titleType') or {}).get('id', '')
            return TYPE_MAP_REVERSE.get(name, ReleaseType.unknown)
        return ReleaseType.unknown

    async def url(self, id):
        if id:
            return f'{WEBSITE_BASE}/title/{id}'
        return ''

    async def year(self, id):
        if id:
            info = await self._get_info(id)
            return str((info.get('releaseYear') or {}).get('year', ''))
        return ''


class _ImdbSearchResult(common.SearchResult):
    def __init__(self, *, imdb_api, info=None, cast=None, countries=None,
                 directors=None, genres=None, id=None, poster=None, summary=None, title=None,
                 title_english=None, title_original=None, type=None, url=None,
                 year=None):
        info = info or {}
        self._imdb_api = imdb_api
        id = id or self._get_id(info)
        super().__init__(
            cast=cast or functools.partial(imdb_api.cast, id),
            countries=countries or functools.partial(imdb_api.countries, id),
            directors=directors or functools.partial(imdb_api.directors, id),
            genres=genres or functools.partial(imdb_api.genres, id),
            id=id,
            poster=functools.partial(imdb_api.poster, id),
            summary=summary or self._get_summary(info),
            title=title or self._get_title(info),
            title_english=title_english or functools.partial(imdb_api.title_english, id),
            title_original=title_original or functools.partial(imdb_api.title_original, id),
            type=type or self._get_type(info),
            url=url or self._get_url(info),
            year=year or self._get_year(info),
        )

    def _get_id(self, info):
        return info.get('id', '')

    def _get_summary(self, info):
        return ((info.get('plot') or {}).get('plotText') or {}).get('plainText', '')

    def _get_title(self, info):
        return (info.get('titleText') or {}).get('text', '')

    def _get_type(self, info):
        name = (info.get('titleType') or {}).get('id', '')
        return TYPE_MAP_REVERSE.get(name, ReleaseType.unknown)

    def _get_url(self, info):
        id = self._get_id(info)
        if id:
            return f'{WEBSITE_BASE}/title/{id}'
        return ''

    def _get_year(self, info):
        return (info.get('releaseYear') or {}).get('year', '')



from base64 import b64decode

from ... import errors
from .. import http
from . import base

import logging  # isort:skip
_log = logging.getLogger(__name__)


class PredbnetApi(base.PredbApiBase):
    name = 'predbnet'
    label = 'PreDB.net'

    default_config = {}

    _url_base = b64decode('cHJlZGIubmV0').decode('ascii')
    _search_url = f'https://api.{_url_base}'

    async def _search(self, query):
        params = self._get_params(query.keywords, query.group)
        return await self._request_all_pages(params)

    def _get_params(self, keywords, group):
        return {
            'type': 'search',
            'q': ' '.join(
                kw_sanitized
                for kw in keywords
                if (kw_sanitized := str(kw).lower().strip())
            ),
            'group': group,
            'limit': self._max_results_per_page,
        }

    # We could request 1000 pages, but we are limited to 30 requests in 60 seconds.
    _max_pages = 30
    _max_results_per_page = 100

    async def _request_all_pages(self, params):
        combined_results = []
        page = 1
        while page <= self._max_pages:
            results, next_page = await self._request_page(params, page)
            combined_results.extend(results)

            if next_page < 0:
                # Negative next page means there are no more pages
                break
            else:
                page += 1

        return combined_results

    async def _request_page(self, params, page):
        params = {**params, 'page': page}
        _log.debug('%s search: %r, %r', self.label, self._search_url, params)
        response = (await http.get(self._search_url, params=params, cache=True, raise_on_error_status=False)).json()

        if response.get('data'):
            # We found at least one release
            results = tuple(result['release'] for result in response['data'])

        elif (
                response.get('status') != 'success'
                and response.get('message')
                and 'no results' not in response['message'].lower()
        ):
            # Report error from API
            raise errors.RequestError(f'{self.label}: {response["message"]}')

        else:
            results = ()

        _log.debug('Found %d results', len(results))

        # Is there another page of results?
        if len(results) >= self._max_results_per_page:
            next_page = page + 1
        else:
            next_page = -1

        return results, next_page

    async def _release_files(self, release_name):
        raise NotImplementedError()



"""
Concrete :class:`~.TrackerJobsBase` subclass for CBR
"""

import functools
import io
import re

from ... import jobs, utils
from ..base import TrackerJobsBase

import logging  # isort:skip
_log = logging.getLogger(__name__)


class CbrTrackerJobs(TrackerJobsBase):

    # Only generate mediainfo/bdinfo reports and screenshots for the first/main video file/playlist.
    document_all_videos = False

    release_name_english_title_before_original = False

    image_host_config = {
        'common': {'thumb_width': 400},
    }

    @functools.cached_property
    def jobs_before_upload(self):
        # NOTE: Keep in mind that the order of jobs is important for
        #       isolated_jobs: The final job is the overall result, so if
        #       upload_screenshots_job is listed after description_job,
        #       --only-description is going to print the list of uploaded
        #       screenshot URLs.
        return (
            # Interactive jobs
            self.playlists_job,
            self.tmdb_job,
            self.imdb_job,
            self.release_name_job,
            self.category_job,
            self.season_job,
            self.episode_job,
            self.type_job,
            self.quality_job,

            # Background jobs
            self.create_torrent_job,
            self.mediainfo_job,
            self.bdinfo_job,
            self.nfo_job,
            self.screenshots_job,
            self.upload_screenshots_job,
            self.description_job,
            self.rules_job,
        )

    @functools.cached_property
    def isolated_jobs(self):
        """
        Sequence of job attribute names (e.g. "description_job") that were
        singled out by the user, e.g. with --only-description
        """
        if self.options.get('only_description', False):
            return self.get_job_and_dependencies(self.description_job)
        elif self.options.get('only_title', False):
            return self.get_job_and_dependencies(
                self.release_name_job,
                # `release_name_job` doesn't depend on `imdb_job` (or any other
                # webdb), but we want the correct name, year, etc in the release
                # name.
                self.imdb_job,
                self.tmdb_job,
            )

        # Activate all jobs in jobs_before/after_upload
        return ()

    @functools.cached_property
    def season_job(self):
        return jobs.dialog.TextFieldJob(
            name=self.get_job_name('season'),
            label='Season',
            precondition=self.make_precondition('season_job'),
            prejobs=(
                self.release_name_job,
                self.category_job,
            ),
            hidden=self.season_job_is_hidden,
            text=self.autodetect_season,
            normalizer=self.normalize_season,
            validator=self.validate_season,
            finish_on_success=True,
            **self.common_job_args(),
        )

    def season_job_is_hidden(self):
        return self.release_name.type not in (
            utils.release.ReleaseType.season,
            utils.release.ReleaseType.episode,
        )

    async def autodetect_season(self):
        if self.season_job_is_hidden():
            return ''

        _log.debug('Autodetected season: type: %r', self.release_name.type)
        if self.release_name.only_season:
            _log.debug('Autodetected season: only_season: %r', self.release_name.only_season)
            return self.release_name.only_season

        return None

    def normalize_season(self, text):
        return text.strip()

    def validate_season(self, text):
        if text:
            try:
                season = int(text)
            except ValueError as e:
                raise ValueError('Season must be a number.') from e
            else:
                # NOTE: Season 0 is ok for pilots, specials, etc.
                if not 0 <= season <= 100:
                    raise ValueError('Season is not reasonable.')

    @functools.cached_property
    def episode_job(self):
        return jobs.dialog.TextFieldJob(
            name=self.get_job_name('episode'),
            label='Episode',
            precondition=self.make_precondition('episode_job'),
            prejobs=(
                self.release_name_job,
                self.category_job,
            ),
            hidden=self.episode_job_is_hidden,
            text=self.autodetect_episode,
            normalizer=self.normalize_episode,
            validator=self.validate_episode,
            finish_on_success=True,
            **self.common_job_args(),
        )

    def episode_job_is_hidden(self):
        return self.release_name.type is not utils.release.ReleaseType.episode

    async def autodetect_episode(self):
        # Use episode 0 for season pack (required).
        if self.release_name.type is utils.release.ReleaseType.season:
            return '0'
        elif self.release_name.type is not utils.release.ReleaseType.episode:
            return ''

        # If episodes as flat list contains only one episode (e.g. "S03E04"), we have our
        # episode. If there are multiple episodes (e.g. "S03E00E01"), we don't do any autodetection
        # and the user must enter the episode.
        all_episode_numbers = tuple(
            episode_number
            for episode_numbers in self.release_name.episodes_dict.values()
            for episode_number in episode_numbers
        )
        if len(all_episode_numbers) == 1:
            _log.debug('Autodetected episode: %r', all_episode_numbers[0])
            return all_episode_numbers[0]

    def normalize_episode(self, text):
        return text.strip()

    def validate_episode(self, text):
        if text:
            try:
                episode = int(text)
            except ValueError as e:
                raise ValueError('Episode must be a number.') from e
            else:
                # NOTE: Episode 0 is ok for pilots, specials, etc.
                if not episode >= 0:
                    raise ValueError('Episode is not reasonable.')

    @functools.cached_property
    def category_job(self):
        return jobs.dialog.ChoiceJob(
            name=self.get_job_name('category'),
            label='Category',
            precondition=self.make_precondition('category_job'),
            prejobs=(
                self.release_name_job,
            ),
            autodetect=self.autodetect_category,
            autofinish=self.release_name.type is utils.release.ReleaseType.movie,
            options=(
                ('Filmes', '1'),
                ('Series', '2'),
                ('Animes', '4'),
                # ('Esportes', '8'),
                # ('Jogos', '5'),
                # ('Programas', '9'),
                # ('HQs/Mangas', '10'),
                # ('Livros', '11'),
                # ('Cursos', '12'),
                # ('Revistas', '13'),
            ),
            **self.common_job_args(),
        )

    _autodetect_category_map = {
        'Filmes': lambda release: release.type is utils.release.ReleaseType.movie,
        'Animes': lambda release: release.type in (
            utils.release.ReleaseType.season,
            utils.release.ReleaseType.episode,
        ) and release.service in (
            'CR',
            'HIDI',
            'Bili',
            'AO',  # anime onegai
            'B-Global', 'B Global',
        ),
        'Series': lambda release: release.type in (
            utils.release.ReleaseType.season,
            utils.release.ReleaseType.episode,
        ),
    }

    def autodetect_category(self, _):
        _log.debug('Autodetected category: type: %r', self.release_name.type)
        _log.debug('Autodetected category: service: %r', self.release_name.service)
        for label, is_match in self._autodetect_category_map.items():
            if is_match(self.release_name):
                return label

    @functools.cached_property
    def type_job(self):
        return jobs.dialog.ChoiceJob(
            name=self.get_job_name('type'),
            label='Type',
            precondition=self.make_precondition('type_job'),
            prejobs=(
                self.release_name_job,
            ),
            autodetect=self.autodetect_type,
            autofinish=True,
            options=(
                ('Full Disc', '1'),
                ('Remux', '2'),
                ('Encode', '3'),
                ('WEB-DL', '4'),
                ('WEBRip', '5'),
                ('HDTV', '6'),
            ),
            focused='WEB-DL',
            **self.common_job_args(),
        )

    _autodetect_type_map = {
        'Full Disc': lambda release: release.source in (
            'DVD5', 'DVD9', 'Blu-ray',
        ),
        'Remux': lambda release: 'REMUX' in release.source,
        'Encode': lambda release: release.source in (
            'HD-DVD', 'BluRay',
        ),
        'WEB-DL': lambda release: 'WEB-DL' in release.source,
        'WEBRip': lambda release: 'WEBRip' in release.source,
        'HDTV': lambda release: 'HDTV' in release.source,
    }

    async def autodetect_type(self, _):
        _log.debug('Autodetected type: source: %r', self.release_name.source)
        for label, is_match in self._autodetect_type_map.items():
            if is_match(self.release_name):
                return label

    @functools.cached_property
    def quality_job(self):
        return jobs.dialog.ChoiceJob(
            name=self.get_job_name('quality'),
            label='Quality',
            precondition=self.make_precondition('quality_job'),
            autodetect=self.autodetect_quality,
            autofinish=True,
            options=(
                *self._quality_map.items(),
                ('Other', '10'),
            ),
            focused='Other',
            **self.common_job_args(),
        )

    _quality_map = {
        '4320p': '1',
        '2160p': '2',
        '1080p': '3',
        '1080i': '4',
        '720p': '5',
        '576p': '6',
        '576i': '7',
        '480p': '8',
        '480i': '9',
    }

    async def autodetect_quality(self, _):
        resolution = utils.mediainfo.video.get_resolution(self.content_path)
        _log.debug('Autodetected quality: %s', resolution)
        if resolution in self._quality_map:
            return resolution

    def translate_video_format(video_format, release_name):
        """
        AVC/HEVC for REMUX/BDMV
        Add DUAL (audio) to the very end (not for BDMV)
        eg: H.264 DUAL-NoGrp
        """
        video_codec = video_format
        if re.search(r'(?:REMUX|Blu\-ray)', release_name.source, flags=re.IGNORECASE):
            if re.search(r'^(?:x|H\.?)264$', video_format, flags=re.IGNORECASE):
                video_codec = 'AVC'
            elif re.search(r'^(?:x|H\.?)265$', video_format, flags=re.IGNORECASE):
                video_codec = 'HEVC'

        if release_name.source != 'Blu-ray' and release_name.has_dual_audio:
            # only tag DUAL if it has Portuguese
            audio_languages = utils.mediainfo.audio.get_audio_languages(
                release_name.path,
                exclude_commentary=True,
            )
            if any(language == 'pt' for language in audio_languages):
                return f'{video_codec} DUAL'

        return video_codec

    def translate_audio_format(audio_format, release_name):
        """
        Remove space between channels for: DD, DDP, AAC
        Move Atmos after channels
        eg: AAC2.0 / DDP5.1 / DDP5.1 Atmos / TrueHD 5.1 Atmos
        """
        channels = utils.mediainfo.audio.get_audio_channels(release_name.path)
        match = re.search(r'(DDP?|AAC|TrueHD)(?: (Atmos))?', audio_format, flags=re.IGNORECASE)

        if match:
            # TrueHD has space (for some reason)
            if match.group(1) == 'TrueHD':
                audio_codec = ' '.join((match.group(1), channels))
            else:
                audio_codec = ''.join((match.group(1), channels))

            return ' '.join((audio_codec, match.group(2) or '')).rstrip()

        # just append channel for the rest, no changes
        return ' '.join((audio_format, channels))

    def translate_audio_channels(audio_channels, release_name):
        # Channel is included by translate_audio_format().
        return ''

    release_name_translation = {
        'source': {
            re.compile(r'(?i:remux)'): 'REMUX',
            re.compile(r'(?i:hybrid)'): 'Hybrid',
            re.compile(r'(?i:bd(?:25|50|66|100))'): 'Blu-ray',
            # Remove UHD if not REMUX/BDMV
            re.compile(r'(?i:uhd (?!.*(Blu-ray|remux)))'): '',
        },
        'video_format': translate_video_format,
        'hdr_format': {
            # HDR10 -> HDR / HDR10+ -> HDR10+
            re.compile(r'(?i:HDR10)(?!\+)'): 'HDR',
        },
        'edition': {
            # DUAL is moved to translate_video_format
            re.compile(r'(?i:dual audio)'): '',
        },
        'group': {
            re.compile(r'^NOGROUP$'): 'NoGroup',
        },
        'audio_format': translate_audio_format,
        'audio_channels': translate_audio_channels,
    }

    @functools.cached_property
    def description_job(self):
        return jobs.dialog.TextFieldJob(
            name=self.get_job_name('description'),
            label='Description',
            precondition=self.make_precondition('description_job'),
            prejobs=(
                self.playlists_job,
                self.mediainfo_job,
                self.bdinfo_job,
                self.screenshots_job,
                self.upload_screenshots_job,
            ),
            text=self.generate_description,
            hidden=True,
            finish_on_success=True,
            read_only=True,
            **self.common_job_args(ignore_cache=True),
        )

    def generate_description(self):
        screenshots = self.generate_description_screenshots()
        parts = [
            (
                f'[center]{screenshots}[/center]'
                if screenshots else
                ''
            ),
        ]

        if promo := self.generate_promotion_bbcode(format='[right][size=1]{message}[/size][/right]'):
            parts.append(promo)

        return '\n\n'.join(part for part in parts if part)

    def generate_description_screenshots(self):
        assert self.upload_screenshots_job.is_finished

        thumb_width = self.image_host_config['common']['thumb_width']
        return ' '.join(
            f'[url={screenshot}][img={thumb_width}]{screenshot}[/img][/url]'
            for screenshot in self.upload_screenshots_job.uploaded_images
        )

    @property
    def post_data(self):
        return {
            'api_token': self.options['apikey'],
            'name': self.get_job_output(self.release_name_job, slice=0),
            'category_id': self.get_job_attribute(self.category_job, 'choice'),
            'type_id': self.get_job_attribute(self.type_job, 'choice'),
            'resolution_id': self.get_job_attribute(self.quality_job, 'choice'),
            'season_number': self.get_job_output(self.season_job, slice=0),
            'episode_number': self.get_job_output(self.episode_job, slice=0),
            'sd': self.post_data_sd,
            'tmdb': self.post_data_tmdb_id,
            'imdb': self.post_data_imdb_id,
            'description': self.get_job_output(self.description_job, slice=0),
            'mediainfo': self.post_data_mediainfo,
            'bdinfo': self.post_data_bdinfo,
            'anonymous': '1' if self.options['anonymous'] else '0',
            'personal_release': '1' if self.options['personal_rip'] else '0',
            'mod_queue_opt_in': '1' if self.options['queue'] else '0',
            # next fields are required
            'tvdb': '0',
            'mal': '0',
            'igdb': '0',
            'stream': '0',  # stream optimized
        }

    @functools.cached_property
    def post_data_sd(self):
        # The SD flag. (0 = Not Standard Definition, 1 = Standard Definition).
        # Default is 0
        try:
            height = int(self.release_name.resolution[:-1])
        except ValueError:
            return '0'
        else:
            return '1' if height < 720 else '0'

    @functools.cached_property
    def post_data_tmdb_id(self):
        # TMDb ID may be 0 for non-existing shows
        if self.tmdb_job.output:
            return self.get_job_output(self.tmdb_job, slice=0).split('/')[1]

        return '0'

    @functools.cached_property
    def post_data_imdb_id(self):
        # numbers only
        if self.imdb_job.output:
            imdb_id = self.get_job_output(self.imdb_job, slice=0)
            match = re.search(r'^(?:tt|)(\d+)$', imdb_id)
            if match:
                return match.group(1)

        return '0'

    @functools.cached_property
    def post_data_nfo(self):
        if nfo := self.read_nfo(strip=True):
            return io.BytesIO(bytes(nfo, 'utf-8'))

    @functools.cached_property
    def post_data_mediainfo(self):
        if not self.is_bdmv_release:
            return self.get_job_output(self.mediainfo_job, slice=0)

    @functools.cached_property
    def post_data_bdinfo(self):
        if self.is_bdmv_release:
            return self.get_job_attribute(self.bdinfo_job, 'quick_summaries')[0]

    @functools.cached_property
    def post_files(self):
        files = {
            'torrent': {
                'file': self.torrent_filepath,
                'mimetype': 'application/x-bittorrent',
            }
        }

        if self.post_data_nfo:
            files['nfo'] = {
                'file': self.post_data_nfo,
                'filename': 'nfo',
                'mimetype': 'text/plain',
            }

        return files



from ..base import rules


class FldFhdOnly(rules.FhdOnly):
    pass


class FldBannedGroup(rules.BannedGroup):

    banned_groups = {
        '4K4U',
        'AOC',
        'C4K',
        'CRUCiBLE',
        'd3g',
        'EASports',
        'FGT',
        'MeGusta',
        'MezRips',
        'nikt0',
        'ProRes',
        'RARBG',
        'ReaLHD',
        'SasukeducK',
        'Sicario',
        'TEKNO3D',
        'Telly',
        'tigole',
        'TOMMY',
        'WKS',
        'x0r',
        'YIFY',
    }



"""
Concrete :class:`~.TrackerJobsBase` subclass for RFX
"""

import functools
import io
import re

from ... import jobs, utils
from ..base import TrackerJobsBase

import logging  # isort:skip
_log = logging.getLogger(__name__)


class RfxTrackerJobs(TrackerJobsBase):

    # Only generate mediainfo/bdinfo reports and screenshots for the first/main video file/playlist.
    document_all_videos = False

    release_name_english_title_before_original = False

    image_host_config = {
        'common': {'thumb_width': 350},
    }

    @functools.cached_property
    def jobs_before_upload(self):
        # NOTE: Keep in mind that the order of jobs is important for
        #       isolated_jobs: The final job is the overall result, so if
        #       upload_screenshots_job is listed after description_job,
        #       --only-description is going to print the list of uploaded
        #       screenshot URLs.
        return (
            # Interactive jobs
            self.playlists_job,
            self.tmdb_job,
            self.imdb_job,
            self.release_name_job,
            self.type_job,
            self.resolution_job,

            # Background jobs
            self.create_torrent_job,
            self.mediainfo_job,
            self.bdinfo_job,
            self.nfo_job,
            self.screenshots_job,
            self.upload_screenshots_job,
            self.description_job,
        )

    @functools.cached_property
    def isolated_jobs(self):
        """
        Sequence of job attribute names (e.g. "description_job") that were
        singled out by the user, e.g. with --only-description
        """
        if self.options.get('only_description', False):
            return self.get_job_and_dependencies(self.description_job)
        elif self.options.get('only_title', False):
            return self.get_job_and_dependencies(
                self.release_name_job,
                # `release_name_job` doesn't depend on `imdb_job` (or any other webdb), but we want
                # the correct name, year, etc in the release name.
                self.imdb_job,
            )

        # Activate all jobs in jobs_before/after_upload
        return ()

    @functools.cached_property
    def type_job(self):
        return jobs.dialog.ChoiceJob(
            name=self.get_job_name('type'),
            label='Type',
            precondition=self.make_precondition('type_job'),
            prejobs=(
                self.release_name_job,
            ),
            autodetect=self.autodetect_type,
            autofinish=True,
            options=(
                ('Full Disc', '43'),
                ('Remux', '40'),
                ('Encode', '41'),
                ('WEB-DL', '42'),
                ('WEBRip', '45'),
                ('HDTV', '35'),
            ),
            focused='Encode',
            **self.common_job_args(),
        )

    _autodetect_type_map = {
        'Remux': lambda release: bool(re.search(r'(?i:Remux)', release.source)),
        'Full Disc': lambda release: bool(re.search(r'^(?i:DVD|BD)', release.source)),
        'Encode': lambda release: bool(re.search(r'^(?i:BluRay|HD-?DVD)', release.source)),
        'WEB-DL': lambda release: bool(re.search(r'^(?i:WEB-DL)', release.source)),
        'WEBRip': lambda release: bool(re.search(r'^(?i:WEBRip)', release.source)),
        'HDTV': lambda release: bool(re.search(r'^(?i:HD-?TV)', release.source)),
    }

    async def autodetect_type(self, _):
        _log.debug('Autodetected type: source: %r', self.release_name.source)
        for label, is_match in self._autodetect_type_map.items():
            if is_match(self.release_name):
                return label

    @functools.cached_property
    def resolution_job(self):
        return jobs.dialog.ChoiceJob(
            name=self.get_job_name('resolution'),
            label='Resolution',
            precondition=self.make_precondition('resolution_job'),
            autodetect=self.autodetect_resolution,
            autofinish=True,
            options=(
                *self._resolution_map.items(),
                ('Other', '10'),
            ),
            focused='Other',
            **self.common_job_args(),
        )

    _resolution_map = {
        '4320p': '1',
        '2160p': '2',
        '1080p': '3',
        '1080i': '4',
        '720p': '5',
        '576p': '6',
        '576i': '7',
        '540p': '11',
        '480p': '8',
        '480i': '9',
    }

    async def autodetect_resolution(self, _):
        resolution = utils.mediainfo.video.get_resolution(self.content_path)
        _log.debug('Autodetected resolution: %s', resolution)
        if resolution in self._resolution_map:
            return resolution

    @functools.cached_property
    def description_job(self):
        return jobs.dialog.TextFieldJob(
            name=self.get_job_name('description'),
            label='Description',
            precondition=self.make_precondition('description_job'),
            prejobs=(
                self.screenshots_job,
                self.upload_screenshots_job,
            ),
            text=self.generate_description,
            hidden=True,
            finish_on_success=True,
            read_only=True,
            **self.common_job_args(ignore_cache=True),
        )

    def generate_description(self):
        screenshots = self.generate_description_screenshots()
        parts = [
            (
                f'[center]{screenshots}[/center]'
                if screenshots else
                ''
            ),
        ]

        if promo := self.generate_promotion_bbcode(format='[right][size=1]{message}[/size][/right]'):
            parts.append(promo)

        return '\n\n'.join(part for part in parts if part)

    def generate_description_screenshots(self):
        assert self.upload_screenshots_job.is_finished
        return utils.bbcode.screenshots_grid(
            screenshots=self.upload_screenshots_job.uploaded_images,
            columns=(2, 3),
            horizontal_spacer=' ',
            vertical_spacer='\n',
        )

    @property
    def post_data(self):
        return {
            'name': self.get_job_output(self.release_name_job, slice=0),
            'description': self.get_job_output(self.description_job, slice=0),
            'mediainfo': self.post_data_mediainfo,
            'bdinfo': self.post_data_bdinfo,
            'category_id': '1',
            'type_id': self.get_job_attribute(self.type_job, 'choice'),
            'resolution_id': self.get_job_attribute(self.resolution_job, 'choice'),
            'tmdb': self.post_data_tmdb_id,
            'imdb': self.post_data_imdb_id,
            'anonymous': '1' if self.options['anonymous'] else '0',
            'personal_release': '1' if self.options['personal_release'] else '0',
        }

    @functools.cached_property
    def post_data_tmdb_id(self):
        return self.get_job_output(self.tmdb_job, slice=0).split('/')[1]

    @functools.cached_property
    def post_data_imdb_id(self):
        return self.get_job_output(self.imdb_job, slice=0)

    @functools.cached_property
    def post_data_mediainfo(self):
        if not self.is_bdmv_release:
            return self.get_job_output(self.mediainfo_job, slice=0)

    @functools.cached_property
    def post_data_bdinfo(self):
        if self.is_bdmv_release:
            return self.get_job_attribute(self.bdinfo_job, 'quick_summaries')[0]

    @functools.cached_property
    def post_files(self):
        files = {
            'torrent': {
                'file': self.torrent_filepath,
                'mimetype': 'application/octet-stream',
            },
        }

        if nfo := self.read_nfo(strip=True):
            files['nfo'] = {
                'file': io.BytesIO(nfo.encode('utf-8')),
                'mimetype': 'text/plain',
            }

        return files
